/*
 *        spChunkP.h
 */

#ifndef __SPCHUNKP_H
#define __SPCHUNKP_H

#include <sp/spChunk.h>

#ifdef __cplusplus
extern "C" {
#endif

#define SP_CHUNK_TYPE_ANY ""    
#define SP_CHUNK_TYPE_TOP "!TOP"    

typedef spChunkSize (*spChunkReadFunc)(spChunk *parent, long depth, void *chunk, int swap, FILE *fp);
typedef spChunkSize (*spChunkWriteFunc)(spChunk *chunk, long depth, spBool rewrite_flag, int swap, FILE *fp);
typedef spBool (*spChunkInitFunc)(spChunk *parent, void *chunk, unsigned long options);
typedef spBool (*spChunkFreeFunc)(void *chunk);
typedef spChunkSize (*spChunkUpdateContentSizeFunc)(void *chunk);

typedef spChunkSize (*spChunkCopyFunc)(spChunk *dest_chunk, spChunk *src_chunk, long depth);
    
typedef struct _spChunkInfoTable {
    char type[5];
    char parent_type[5];
    spBool mandatory_flag;
    unsigned long options;
    int offset;
    int size;
    spChunkSize initial_content_size;
    spChunkReadFunc read_func;
    spChunkWriteFunc write_func;
    spChunkInitFunc init_func;
    spChunkFreeFunc free_func;
    spBool free_self;
    spChunkUpdateContentSizeFunc update_content_size_func;

    spChunkCopyFunc copy_func;
} spChunkInfoTable;

typedef spChunkSize (*spChunkHeaderGetSizeFunc)(spChunkHeader *header);
typedef spChunkSize (*spChunkGetSizeFunc)(spChunk *chunk);
typedef spBool (*spChunkSetSizeFunc)(spChunk *chunk, spChunkSize size);
typedef spChunkSize (*spChunkReadHeaderFunc)(spChunk *parent, spChunkHeader *header, int swap, FILE *fp);
typedef spChunkSize (*spChunkReadHeaderRestFunc)(spChunk *parent, spChunkHeader *o_header, spChunkHeader *i_header, unsigned long options, int swap, FILE *fp);
typedef spChunkSize (*spChunkSkipFunc)(spChunk *parent, spChunkHeader *header, spChunkSize remain_size, int swap, FILE *fp);

typedef spChunkSize (*spChunkWriteHeaderFunc)(spChunk *chunk, spChunkHeader *header, int swap, FILE *fp);
typedef spChunkSize (*spChunkWriteHeaderRestFunc)(spChunk *chunk, spChunkHeader *header, unsigned long options, int swap, FILE *fp);
typedef spChunkSize (*spChunkWriteContentRestFunc)(spChunk *chunk, long depth, spChunkSize remain_size, spBool rewrite_flag, int swap, FILE *fp);

typedef spChunkSize (*spChunkCopyHeaderFunc)(spChunkHeader *dest_header, spChunkHeader *src_header);
    
typedef struct _spChunkFileSpec {
    int header_size;
    int default_struct_size;
    
    spChunkInfoTable *info_list;
    int num_list;

    spChunkInitFunc init_chunk_func;
    spChunkHeaderGetSizeFunc get_header_size_from_header_func;
    spChunkHeaderGetSizeFunc get_content_size_from_header_func;
    spChunkGetSizeFunc get_header_size_func;
    spChunkGetSizeFunc get_content_size_func;
    spChunkSetSizeFunc set_content_size_func;
    spChunkGetSizeFunc get_content_margin_size_func;
    spChunkSetSizeFunc set_content_margin_size_func;
    spChunkReadHeaderFunc read_header_func;
    spChunkReadHeaderRestFunc read_header_rest_func;
    spChunkSkipFunc skip_func;

    spChunkWriteHeaderFunc write_header_func;
    spChunkWriteHeaderRestFunc write_header_rest_func;
    spChunkWriteContentRestFunc write_content_rest_func;

    spChunkCopyHeaderFunc copy_header_func;
} spChunkFileSpec;
    
extern void spPropagateChunkContentSize(spChunkFileSpec *spec, spChunk *parent, spChunkSize size_diff);
    
extern spChunkSize spCreateChildChunk(spChunkFileSpec *spec, spChunk *chunk,
                                      spBool propagate_size, spBool recursive_flag, spChunkUpdateFunc update_func, void *data);
extern spChunk *spCreateChunkFromInfoTable(spChunkFileSpec *spec, spChunk *parent, spChunk *insert_target, spChunkInfoTable *chunk_info,
                                           spBool propagate_size, spBool recursive_flag, spChunkUpdateFunc update_func, void *data);
extern spChunk *spCreateChunk(spChunkFileSpec *spec, spChunk *parent, spChunk *insert_target, const char *type,
                              spBool propagate_size, spBool recursive_flag, spChunkUpdateFunc update_func, void *data);

extern spBool spSetChunkContentSize(spChunkFileSpec *spec, spChunk *chunk, spChunkSize size, spBool propagate_size);
    
extern spChunkSize spUpdateChunkContentSize(spChunkFileSpec *spec, spChunk *chunk, spBool propagate_size);
extern spChunkSize spUpdateChunk(spChunkFileSpec *spec, spChunk *chunk, spChunkUpdateFunc update_func, void *data);
extern spChunkSize spUpdateChildChunk(spChunkFileSpec *spec, spChunk *chunk, spChunkUpdateFunc update_func, void *data);
    
extern spChunkSize spReadChunk(spChunkFileSpec *spec, spChunk *parent, spChunkSize remain_size, long depth, int swap, FILE *fp);
extern spChunkSize spReadChildChunk(spChunkFileSpec *spec, spChunk *parent, long max_count, spChunkSize remain_size, long depth, int swap, FILE *fp);
extern spBool spFreeChunk(spChunkFileSpec *spec, spChunk *chunk);
extern spChunkInfoTable *spFindChunkInfoTable(spChunkInfoTable *info_list, int num_list,
                                              spChunkHeader *parent_header, spChunkHeader *header);
extern spChunkInfoTable *spFindChildChunkInfoTable(spChunkInfoTable *info_list, int num_list,
                                                   spChunkHeader *parent_header, int *index);

extern spChunkSize spWriteChunk(spChunkFileSpec *spec, spChunk *chunk, long depth,
                                spBool recursive_flag, spBool rewrite_flag, int swap, spBool *paused, FILE *fp);
extern spChunkSize spWriteChildChunk(spChunkFileSpec *spec, spChunk *chunk, long depth,
                                     spBool recursive_flag, spBool rewrite_flag, int swap, spChunk **paused_chunk, FILE *fp);
    
extern spChunkSize spCopyChunk(spChunkFileSpec *spec, spChunk *dest_parent, spChunk *src_chunk, long depth);
extern spChunkSize spCopyChildChunk(spChunkFileSpec *spec, spChunk *dest_parent, spChunk *src_parent, long depth);

#ifdef __cplusplus
}  /* Close scope of 'extern "C"' declaration */
#endif

#endif /* __SPCHUNKP_H */
