/*
 *        spChunk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>

#include <sp/spChunkP.h>

spChunk *spFindChunk(spChunk *chunk, const char *type, const char *parent_type)
{
    spChunk *child;

    /*if (strnone(type)) return NULL;*/

    spDebug(100, "spFindChunk", "type = %s\n", type);
    
    while (chunk != NULL) {
        if (((type != NULL && strlen(type) > 0 && strneq(chunk->header.type, type, strlen(type)))
             || (chunk->parent != NULL && strnone(chunk->header.type) && strnone(type)))
            && (strnone(parent_type)
                || chunk->parent == NULL
                || strneq(chunk->parent->header.type, parent_type, strlen(parent_type)))) {
            spDebug(100, "spFindChunk", "found: %s\n", type);
            return chunk;
        } else if (chunk->child != NULL) {
            if ((child = spFindChunk(chunk->child, type, parent_type)) != NULL) {
                spDebug(100, "spFindChunk", "child found: %s\n", type);
                return child;
            }
        }

        chunk = chunk->next;
    }

    spDebug(10, "spFindChunk", "**** not found: %s ****\n", type);
    
    return NULL;
}

spChunk *spFindChildChunk(spChunk *top, const char *type, const char *parent_type)
{
    if (top == NULL || strnone(type)) return NULL;
    
    return spFindChunk(top->child, type, parent_type);
}

void spPropagateChunkContentSize(spChunkFileSpec *spec, spChunk *parent, spChunkSize size_diff)
{
    spChunkSize parent_content_size;
    spChunkSize margin_size;

    if (size_diff == 0) return;
        
    spDebug(100, "spPropagateChunkContentSize", "size_diff = %lu\n", (unsigned long)size_diff);
    
    while (parent != NULL) {
        parent_content_size = spec->get_content_size_func(parent);
        spDebug(100, "spPropagateChunkContentSize", "content_size = %lu, type = %c%c%c%c\n",
                (unsigned long)parent_content_size, parent->header.type[0], parent->header.type[1],
                parent->header.type[2], parent->header.type[3]);

        if (spec->get_content_margin_size_func != NULL) {
            margin_size = spec->get_content_margin_size_func(parent);
            if (margin_size >= size_diff) {
                margin_size -= size_diff;
                spec->set_content_margin_size_func(parent, margin_size);
                /* size_diff is absorbed, so stop propagating */
                break;
            } else {
                size_diff -= margin_size;
                spec->set_content_margin_size_func(parent, 0);
            }
        }

        
        spec->set_content_size_func(parent, parent_content_size + size_diff);

        parent = parent->parent;
    }

    return;
}

spChunkSize spCreateChildChunk(spChunkFileSpec *spec, spChunk *chunk,
                               spBool propagate_size, spBool recursive_flag, spChunkUpdateFunc update_func, void *data)
{
    int i;
    spChunkSize orig_size;
    spChunkSize total_size;
    spChunk *child;
    spChunkInfoTable *child_chunk_info;

    if (chunk == NULL) return 0;

    spDebug(80, "spCreateChildChunk", "chunk type = %c%c%c%c, spec->num_list = %d, propagate_size = %d\n",
            chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3],
            spec->num_list, propagate_size);
    
    orig_size = spec->get_content_size_func(chunk);
    total_size = orig_size;
    spDebug(80, "spCreateChildChunk", "before loop: total_size = %lu\n", (unsigned long)total_size);
    
    for (i = 0;;) {
        spDebug(80, "spCreateChildChunk", "i = %d\n", i);
        
        child_chunk_info = spFindChildChunkInfoTable(spec->info_list, spec->num_list,
                                                     &chunk->header, &i);
        if (child_chunk_info == NULL) {
            spDebug(80, "spCreateChildChunk", "spFindChildChunkInfoTable returns NULL\n");
            break;
        }
        spDebug(80, "spCreateChildChunk", "child_chunk_info found: i = %d\n", i);

        if (child_chunk_info->mandatory_flag == SP_TRUE) {
            spDebug(80, "spCreateChildChunk", "mandatory child_chunk_info: type = %s\n", child_chunk_info->type);
            child = spCreateChunkFromInfoTable(spec, chunk, NULL, child_chunk_info, SP_FALSE, recursive_flag, update_func, data);
            if (child != NULL) {
                total_size += spec->get_header_size_func(child);
                total_size += spec->get_content_size_func(child);
            }
        }
        ++i;
        spDebug(80, "spCreateChildChunk", "updated: i = %d\n", i);
    }

    spDebug(80, "spCreateChildChunk", "after loop: total_size = %lu\n", (unsigned long)total_size);
    
    if (propagate_size) {
        spec->set_content_size_func(chunk, orig_size);
    
        spDebug(80, "spCreateChildChunk", "call spPropagateChunkContentSize\n");
        spPropagateChunkContentSize(spec, chunk, total_size - orig_size);
    } else {
        spec->set_content_size_func(chunk, total_size);
    }

    spDebug(80, "spCreateChildChunk", "done: total_size = %lu\n", (unsigned long)total_size);
    
    return total_size;
}

spChunk *spCreateChunkFromInfoTable(spChunkFileSpec *spec, spChunk *parent, spChunk *insert_target, spChunkInfoTable *chunk_info,
                                    spBool propagate_size, spBool recursive_flag, spChunkUpdateFunc update_func, void *data)
{
    char *ptr;
    int size;
    int offset;
    unsigned long options;
    spChunkSize tsize;
    spChunkSize content_size;
    spChunk *prev_chunk;
    spChunk *next_chunk;
    spChunk *chunk;

    size = spec->default_struct_size;
    spDebug(50, "spCreateChunkFromInfoTable", "sizeof chunk = %d\n", size);

    if (chunk_info != NULL) {
        spDebug(50, "spCreateChunkFromInfoTable", "chunk_info->size = %d, chunk_info->offset = %d\n",
                chunk_info->size, chunk_info->offset);
        if (chunk_info->size > 0) {
            size = chunk_info->size;
        }
        offset = chunk_info->offset;
        content_size = chunk_info->initial_content_size;
        options = chunk_info->options;
    } else {
        offset = 0;
        content_size = 0;
        options = 0;
    }
    spDebug(50, "spCreateChunkFromInfoTable", "offset = %d, size = %d, content_size = %lu\n", offset, size, (unsigned long)content_size);

    ptr = xalloc(size, char);
    memset(ptr, 0, size);
    spDebug(50, "spCreateChunkFromInfoTable", "alloc done\n");
                
    chunk = (spChunk *)ptr;

    if (parent == NULL && insert_target != NULL) {
        parent = insert_target->parent;
    }
    
    if (parent != NULL) {
        if (parent->child == NULL) {
            spDebug(50, "spCreateChunkFromInfoTable", "first child\n");
            parent->child = chunk;
            prev_chunk = NULL;
            next_chunk = NULL;
        } else if (insert_target != NULL) {
            prev_chunk = insert_target->prev;
            next_chunk = insert_target;
            if (prev_chunk == NULL) {
                parent->child = chunk;
            }
        } else {
            spDebug(50, "spCreateChunkFromInfoTable", "non-first child\n");
            prev_chunk = parent->child;
            while (prev_chunk->next != NULL) {
                prev_chunk = prev_chunk->next;
            }
            prev_chunk->next = chunk;
            next_chunk = NULL;
        }
        chunk->prev = prev_chunk;
        chunk->next = next_chunk;

        if (offset > 0) {
            spChunk **chunkp;
        
            spDebug(50, "spCreateChunkFromInfoTable", "set this to parent member\n");
            chunkp = (spChunk **)((char *)parent + offset);
            if (*chunkp == NULL) {
                *chunkp = chunk;
                spDebug(50, "spCreateChunkFromInfoTable", "setting this to parent member OK\n");
            }
        }
    }

    if (chunk_info != NULL) {
        memcpy(chunk->header.type, chunk_info->type, 4);
        spDebug(50, "spCreateChunkFromInfoTable", "type = %c%c%c%c\n",
                chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3]);
    }

    if (spec->init_chunk_func != NULL) {
        spDebug(50, "spCreateChunkFromInfoTable", "call init_chunk_func\n");
        if (spec->init_chunk_func(parent, chunk, options) == SP_FALSE) {
            spDebug(50, "spCreateChunkFromInfoTable", "init_chunk_func failed\n");
            xfree(ptr);
            return NULL;
        }
    }
    if (chunk_info != NULL && chunk_info->init_func != NULL) {
        if (chunk_info->init_func(parent, chunk, options) == SP_FALSE) {
            spDebug(50, "spCreateChunkFromInfoTable", "init_func failed\n");
            xfree(ptr);
            return NULL;
        }
        spDebug(50, "spCreateChunkFromInfoTable", "chunk init_func done\n");
    }

    if (content_size <= 0 && chunk_info != NULL && chunk_info->update_content_size_func != NULL) {
        if ((tsize = chunk_info->update_content_size_func(chunk)) > 0) {
            content_size = tsize;
            spDebug(50, "spCreateChunkFromInfoTable", "update_content_size_func result = %lu\n", (unsigned long)content_size);
        }
    }
    
    spDebug(50, "spCreateChunkFromInfoTable", "call set_content_size_func content_size = %lu\n", (unsigned long)content_size);
    spec->set_content_size_func(chunk, content_size);
    
    if (update_func != NULL) {
        if ((tsize = update_func(chunk, data)) > 0) {
            content_size = tsize;
            spec->set_content_size_func(chunk, content_size);
            spDebug(50, "spCreateChunkFromInfoTable", "update_func result = %lu\n", (unsigned long)content_size);
        }
    }

    if (recursive_flag == SP_TRUE && chunk_info != NULL) {
        content_size = spCreateChildChunk(spec, chunk, SP_FALSE, SP_TRUE, update_func, data);
    }

    chunk->parent = parent;        /* doing this here is to prevent from propagating */
    
    if (propagate_size) {
        spPropagateChunkContentSize(spec, chunk->parent, content_size + spec->get_header_size_func(chunk));
    }

    spDebug(50, "spCreateChunkFromInfoTable", "done: content_size = %d, type = %c%c%c%c\n",
            (long)content_size, chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3]);
    
    return chunk;
}

spChunk *spCreateChunk(spChunkFileSpec *spec, spChunk *parent, spChunk *insert_target, const char *type,
                       spBool propagate_size, spBool recursive_flag, spChunkUpdateFunc update_func, void *data)
{
    spChunkHeader chunk_header;
    spChunkInfoTable *chunk_info;
    spChunkHeader *parent_header = NULL;

    if (spec == NULL || type == NULL) return NULL;

    spDebug(50, "spCreateChunk", "type = %c%c%c%c\n", type[0], type[1], type[2], type[3]);
    
    if (parent != NULL) {
        parent_header = &parent->header;
    }
    
    memset(&chunk_header, 0, sizeof(spChunkHeader));
    memcpy(chunk_header.type, type, 4);
    
    chunk_info = spFindChunkInfoTable(spec->info_list, spec->num_list, parent_header, &chunk_header);
    
    return spCreateChunkFromInfoTable(spec, parent, insert_target, chunk_info, propagate_size, recursive_flag, update_func, data);
}

spBool spSetChunkContentSize(spChunkFileSpec *spec, spChunk *chunk, spChunkSize size, spBool propagate_size)
{
    spBool flag;
    spChunkSize orig_size;
    spChunkSize size_diff = 0;
    
    if (spec == NULL || chunk == NULL) return SP_FALSE;

    spDebug(100, "spSetChunkContentSize", "%c%c%c%c: size = %lu, propagate_size = %d\n",
            chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3],
            (unsigned long)size, propagate_size);
    
    orig_size = spec->get_content_size_func(chunk);
    spDebug(100, "spSetChunkContentSize", "orig_size = %lu\n", (unsigned long)orig_size);

    if (size == orig_size) {
        return SP_TRUE;
    }

    if (spec->get_content_margin_size_func != NULL) {
        spChunkSize actual_content_size;
        spChunkSize margin_size;
        
        margin_size = spec->get_content_margin_size_func(chunk);
        actual_content_size = orig_size - margin_size;
        if (size > actual_content_size) {
            size_diff = size - actual_content_size;

            if (margin_size >= size_diff) {
                margin_size -= size_diff;
                spec->set_content_margin_size_func(chunk, margin_size);
                /* size_diff is absorbed, so return */
                return SP_TRUE;
            } else {
                size_diff -= margin_size;
                size -= margin_size;
                spec->set_content_margin_size_func(chunk, 0);
            }
        } else {
            margin_size += (actual_content_size - size);
            spec->set_content_margin_size_func(chunk, margin_size);
            return SP_TRUE;
        }
    } else {
        if (size < orig_size) {
            return SP_TRUE;
        }
    }

    flag = spec->set_content_size_func(chunk, size);
    
    if (propagate_size == SP_TRUE && flag == SP_TRUE) {
        if (size_diff == 0) {
            size_diff = size - orig_size;
        }
        spPropagateChunkContentSize(spec, chunk->parent, size_diff);
    }

    spDebug(100, "spSetChunkContentSize", "done: size_diff = %lu\n", (unsigned long)size_diff);
    
    return flag;
}

spChunkSize spUpdateChunkContentSize(spChunkFileSpec *spec, spChunk *chunk, spBool propagate_size)
{
    spChunkSize size = 0;
    spChunkSize orig_size = 0;
    spChunkInfoTable *chunk_info;
    spChunkHeader *parent_header = NULL;

    if (spec == NULL || chunk == NULL) return 0;

    spDebug(100, "spUpdateChunkContentSize", "%c%c%c%c: propagate_size = %d\n",
            chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3],
            propagate_size);
    
    if (propagate_size) {
        orig_size = spec->get_content_size_func(chunk);
        spDebug(100, "spUpdateChunkContentSize", "orig_size = %lu\n", (unsigned long)orig_size);
    }

    if (chunk->parent != NULL) {
        parent_header = &chunk->parent->header;
    }
    
    chunk_info = spFindChunkInfoTable(spec->info_list, spec->num_list, parent_header, &chunk->header);
    
    if (chunk_info != NULL && chunk_info->update_content_size_func != NULL) {
        size = chunk_info->update_content_size_func(chunk);
        if (size > 0) {
            spec->set_content_size_func(chunk, size);
        }
    }

    if (size <= 0) {
        size = spec->get_content_size_func(chunk);
    }
    
    spDebug(100, "spUpdateChunkContentSize", "size = %lu\n", (unsigned long)size);

    if (propagate_size) {
        spPropagateChunkContentSize(spec, chunk->parent, size - orig_size);
    }

    return size;
}

spChunkSize spUpdateChunk(spChunkFileSpec *spec, spChunk *chunk, spChunkUpdateFunc update_func, void *data)
{
    spChunkSize size = 0;
    
    if (update_func != NULL) {
        spChunkSize orig_size = 0;
        
        orig_size = spec->get_content_size_func(chunk);
        
        if ((size = update_func(chunk, data)) > 0) {
            spPropagateChunkContentSize(spec, chunk->parent, size - orig_size);
        } else {
            size = orig_size;
        }
    } else {
        size = spUpdateChunkContentSize(spec, chunk, SP_TRUE);
    }

    return size;
}

spChunk *spGetLastChildChunk(spChunk *parent)
{
    spChunk *chunk;

    chunk = parent->child;
    if (chunk != NULL) {
        while (chunk->next != NULL) {
            chunk = chunk->next;
        }
    }

    return chunk;
}

spChunkSize spUpdateChildChunk(spChunkFileSpec *spec, spChunk *chunk, spChunkUpdateFunc update_func, void *data)
{
    spChunkSize size;
    spChunkSize total_size;
    spChunk *child;
    
    if (chunk == NULL) return 0;

    total_size = 0;
    child = chunk->child;

    while (child != NULL) {
        total_size += spec->get_header_size_func(child);

        if (update_func != NULL) {
            size = update_func(child, data);
        } else {
            size = 0;
        }
        if (size <= 0) {
            size = spec->get_content_size_func(child);
        }
        total_size += size;

        child = child->next;
    }
    
    return total_size;
}

spChunkSize spReadChunk(spChunkFileSpec *spec, spChunk *parent, spChunkSize remain_size, long depth, int swap, FILE *fp)
{
    spChunkSize nread;
    spFileOffset offset;
    spChunkHeader chunk_header;
    unsigned long options;
    spChunkSize initial_remain_size;
    spChunkSize total_nread;
    spChunkInfoTable *chunk_info;
    spChunk *chunk;

    spDebug(50, "spReadChunk", "in: remain_size = %lu\n", (unsigned long)remain_size);
    
    if (remain_size > 0 && remain_size < 8) {
        if (spSeekFile(fp, remain_size, SEEK_CUR) != 0) {
            return 0;
        }
        return remain_size;
    }

    initial_remain_size = remain_size;

    memset(&chunk_header, 0, sizeof(spChunkHeader));

    offset = spTellFile(fp);
    
    if ((nread = spec->read_header_func(parent, &chunk_header, swap, fp)) <= 0) {
        spDebug(10, "spReadChunk", "calling read_header_func failed\n");
        return 0;
    }
    spDebug(50, "spReadChunk", "depth = %ld, nread = %lu\n", depth, (unsigned long)nread);
    
    total_nread = nread;
    if (remain_size > 0) remain_size -= nread;

    chunk_info = spFindChunkInfoTable(spec->info_list, spec->num_list, &parent->header, &chunk_header);
    if (chunk_info != NULL) {
        options = chunk_info->options;
        spDebug(50, "spReadChunk", "options = %lx\n", options);

        if ((options & SP_CHUNK_INFO_OPTION_MARGIN_CHUNK_MASK)
            && spec->set_content_margin_size_func != NULL) {
            spChunkSize margin_size = 0;
                
            if ((chunk = spGetLastChildChunk(parent)) != NULL) {
                spChunkSize orig_margin_size = 0;
                if (spec->get_content_margin_size_func != NULL) {
                    orig_margin_size = spec->get_content_margin_size_func(chunk);
                    margin_size = orig_margin_size;
                    spDebug(50, "spReadChunk", "original content margin size of last chunk (%c%c%c%c) = %lu\n",
                            chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3],
                            (unsigned long)margin_size);
                }
                if (spec->get_content_size_from_header_func != NULL) {
                    margin_size += spec->get_header_size_from_header_func(&chunk_header);
                    margin_size += spec->get_content_size_from_header_func(&chunk_header);
                } else {
                    margin_size += initial_remain_size;
                }
                spDebug(50, "spReadChunk", "updated content margin size = %lu\n", (unsigned long)margin_size);
                spec->set_content_margin_size_func(chunk, margin_size);
                spec->set_content_size_func(chunk, margin_size + spec->get_content_size_func(chunk) - orig_margin_size);

#if 0
                spDebug(50, "spReadChunk", "updated content size = %lu\n", (unsigned long)spec->get_content_size_func(chunk));
#endif
            }

            if (initial_remain_size == 0) {
                remain_size = spec->get_content_size_from_header_func(&chunk_header);
            }

            spDebug(10, "spReadChunk", "remain_size = %lu\n", (unsigned long)remain_size);
            
            if (remain_size > 0) {
                if ((nread = spec->skip_func(parent, &chunk_header, remain_size, swap, fp)) > 0) {
                    total_nread += nread;
                    remain_size -= nread;
                } else {
                    spDebug(10, "spReadChunk", "calling skip_func failed: nread = %lu\n", (unsigned long)nread);
                }
            }
            
            spDebug(10, "spReadChunk", "margin: total_nread = %lu, remain_size = %lu\n",
                    (unsigned long)total_nread, (unsigned long)remain_size);
            
            return total_nread;
        }
    } else {
        options = 0;
    }
    spDebug(50, "spReadChunk", "options = %lx, header_size = %d\n", options, spec->header_size);

    chunk = spCreateChunkFromInfoTable(spec, parent, NULL, chunk_info, SP_FALSE, SP_FALSE, NULL, NULL);
    chunk->offset = offset;
        
    if (options && spec->read_header_rest_func != NULL) {
        if ((nread = spec->read_header_rest_func(parent, &chunk->header, &chunk_header, options, swap, fp)) <= 0) {
            spDebug(10, "spReadChunk", "read_header_rest_func failed: nread = %lu\n", (unsigned long)nread);
        } else {
            total_nread += nread;
            if (remain_size > 0) remain_size -= nread;
            spDebug(50, "spReadChunk", "after read_header_rest_func: nread = %lu, total_nread = %lu\n",
                    (unsigned long)nread, (unsigned long)total_nread);
        }
    } else {
        memcpy(&chunk->header, &chunk_header, spec->header_size);
    }

    spDebug(50, "spReadChunk", "before read %c%c%c%c: nread = %lu, total_nread = %lu, remain_size = %lu\n",
            chunk_header.type[0], chunk_header.type[1], chunk_header.type[2], chunk_header.type[3],
            (unsigned long)nread, (unsigned long)total_nread, (unsigned long)remain_size);
    
    if (nread >= 0 && (initial_remain_size == 0 || remain_size > 0)) {
        if (chunk_info != NULL && chunk_info->read_func != NULL) {
            if ((nread = chunk_info->read_func((spChunk *)parent, depth, chunk, swap, fp)) <= 0) {
                spDebug(10, "spReadChunk", "calling read_func failed: nread = %lu\n", (unsigned long)nread);
            }
        } else {
            if ((nread = spec->skip_func(parent, &chunk->header, remain_size, swap, fp)) < 0) {
                spDebug(10, "spReadChunk", "calling skip_func failed: nread = %lu\n", (unsigned long)nread);
            }
        }

        if (nread > 0) {
            total_nread += nread;
            if (remain_size > 0) remain_size -= nread;
        }
        spDebug(50, "spReadChunk", "after read %c%c%c%c: nread = %lu, total_nread = %lu, remain_size = %lu\n",
                chunk_header.type[0], chunk_header.type[1], chunk_header.type[2], chunk_header.type[3],
                (unsigned long)nread, (unsigned long)total_nread, (unsigned long)remain_size);
    }

    /* 2021/8 added to prevent wrong file detection */
    if (chunk_header.type[0] == NUL || chunk_header.type[1] == NUL
        || chunk_header.type[2] == NUL/* || chunk_header.type[3] == NUL*/) {
        spDebug(10, "spReadChunk", "type includes NULL character, maybe wrong file\n");
        nread = -1;
    }
        
    if (nread >= 0) {
        spDebug(50, "spReadChunk", "%c%c%c%c: total_nread = %lu\n",
                chunk_header.type[0], chunk_header.type[1], chunk_header.type[2], chunk_header.type[3], (unsigned long)total_nread);
    
        return total_nread;
    } else {
        spFreeChunk(spec, chunk);
        return 0;
    }
}

spChunkSize spReadChildChunk(spChunkFileSpec *spec, spChunk *parent, long max_count, spChunkSize remain_size, long depth, int swap, FILE *fp)
{
    long k;
    spChunkSize nread;
    spChunkSize total_nread;
    spChunkSize content_size;

    total_nread = 0;
    content_size = spec->get_content_size_func(parent);
    
    if (remain_size <= 0) {
        remain_size = content_size;
    }
    spDebug(80, "spReadChildChunk", "depth = %ld, remain_size = %lu, content_size = %lu, type = %c%c%c%c\n",
            depth, (unsigned long)remain_size, (unsigned long)content_size, 
            parent->header.type[0], parent->header.type[1], parent->header.type[2], parent->header.type[3]);

    nread = 0;
    k = 0;
    while ((content_size <= 1 || remain_size > 0)
           && (nread = spReadChunk(spec, parent, remain_size, depth + 1, swap, fp)) > 0) {
        spDebug(80, "spReadChildChunk", "in loop (%ld / %ld): nread = %lu, total_nread = %lu / %lu, remain_size = %lu, depth = %ld\n",
                k, max_count, (unsigned long)nread, (unsigned long)total_nread, (unsigned long)content_size, (unsigned long)remain_size, depth);
        
        total_nread += nread;
        if (remain_size <= nread) {
            remain_size = 0;
        } else {
            remain_size -= nread;

            if (remain_size < 8) {
                spDebug(40, "spReadChildChunk", "seek because remain_size (%lu) is small\n", (unsigned long)remain_size);
                if (spSeekFile(fp, remain_size, SEEK_CUR) != 0) {
                    return 0;
                }
                total_nread += remain_size;
                remain_size = 0;
            }
        }

        k++;
        if (max_count > 0 && k >= max_count) {
            break;
        }
    }

    if (nread < 0) {
        return 0;
    }

    spDebug(80, "spReadChildChunk", "done: total_nread = %lu, remain_size = %lu, depth = %ld, type = %c%c%c%c\n",
            (unsigned long)total_nread, (unsigned long)remain_size, depth,
            parent->header.type[0], parent->header.type[1], parent->header.type[2], parent->header.type[3]);
    
    return total_nread;
}

spBool spFreeChunk(spChunkFileSpec *spec, spChunk *chunk)
{
    spChunkInfoTable *chunk_info;
    spChunk *parent_chunk;
    spChunkHeader *parent_header;
    spChunk *child_chunk;
    spChunk *next_child_chunk;
    spChunk **chunkp;
    spBool free_self;

    if (chunk == NULL) return SP_FALSE;

    spDebug(80, "spFreeChunk", "type = %c%c%c%c\n",
            chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3]);

    parent_chunk = chunk->parent;
    child_chunk = chunk->child;

    while (child_chunk != NULL) {
        next_child_chunk = child_chunk->next;

        spDebug(80, "spFreeChunk", "parent = %c%c%c%c, child type = %c%c%c%c\n",
                chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3],
                child_chunk->header.type[0], child_chunk->header.type[1], child_chunk->header.type[2], child_chunk->header.type[3]);
    
        spFreeChunk(spec, child_chunk);
        
        child_chunk = next_child_chunk;
    }

    if (parent_chunk != NULL) {
        parent_header = &parent_chunk->header;
    } else {
        parent_header = NULL;
    }
    
    chunk_info = spFindChunkInfoTable(spec->info_list, spec->num_list, parent_header, &chunk->header);
        
    if (chunk_info != NULL) {
        if (chunk_info->free_func != NULL) {
            chunk_info->free_func(chunk);
        }
        if (parent_chunk != NULL && chunk_info->offset > 0) {
            chunkp = (spChunk **)((char *)parent_chunk + chunk_info->offset);
            *chunkp = NULL;
        }
        free_self = chunk_info->free_self;
    } else {
        free_self = SP_FALSE;
    }

    if (chunk->prev != NULL) {
        chunk->prev->next = chunk->next;
    }
    if (chunk->next != NULL) {
        chunk->next->prev = chunk->prev;
    }
    if (parent_chunk != NULL && parent_chunk->child == chunk) {
        parent_chunk->child = chunk->next;
    }

    if (free_self == SP_TRUE) {
        xfree(chunk);
    }

    spDebug(80, "spFreeChunk", "done\n");
    
    return SP_TRUE;
}

spChunkInfoTable *spFindChunkInfoTable(spChunkInfoTable *info_list, int num_list,
                                       spChunkHeader *parent_header, spChunkHeader *header)
{
    int i;

    if (info_list == NULL || header == NULL) return NULL;
    
    for (i = 0; i < num_list; i++) {
        spDebug(100, "spFindChunkInfoTable", "i = %d\n", i);
        if (((strlen(info_list[i].type) > 0 && strneq(info_list[i].type, header->type, strlen(info_list[i].type)))
             || (parent_header != NULL && strnone(header->type) && strnone(info_list[i].type)))
            && (parent_header == NULL
                || strnone(info_list[i].parent_type)
                || strneq(info_list[i].parent_type, parent_header->type, strlen(info_list[i].parent_type)))) {
            spDebug(100, "spFindChunkInfoTable", "found: i = %d, type = %s\n", i, info_list[i].type);
            return &info_list[i];
        }
    }

    spDebug(100, "spFindChunkInfoTable", "not found: i = %d\n", i);
    
    return NULL;
}

spChunkInfoTable *spFindChildChunkInfoTable(spChunkInfoTable *info_list, int num_list,
                                            spChunkHeader *parent_header, int *index)
{
    int i;

    if (info_list == NULL || parent_header == NULL) return NULL;
    
    spDebug(80, "spFindChildChunkInfoTable", "parent_header->type = %c%c%c%c\n",
            parent_header->type[0], parent_header->type[1], parent_header->type[2], parent_header->type[3]);
    
    for (i = *index; i < num_list; i++) {
        if (info_list[i].mandatory_flag == SP_FALSE) {
            continue;
        }
        spDebug(100, "spFindChildChunkInfoTable", "info_list[%d].type = %s\n", i, info_list[i].type);
        
        if ((strnone(parent_header->type)
             && (strnone(info_list[i].parent_type) || streq(info_list[i].parent_type, SP_CHUNK_TYPE_TOP)))
            || (!strnone(info_list[i].parent_type)
                && strneq(info_list[i].parent_type, parent_header->type, strlen(info_list[i].parent_type)))) {
            *index = i;
            spDebug(100, "spFindChildChunkInfoTable", "found: i = %d, type = %s\n", i, info_list[i].type);
            return &info_list[i];
        }
    }

    spDebug(10, "spFindChildChunkInfoTable", "not found: i = %d\n", i);
    
    return NULL;
}

spChunkSize spWriteChunk(spChunkFileSpec *spec, spChunk *chunk, long depth,
                         spBool recursive_flag, spBool rewrite_flag, int swap, spBool *paused, FILE *fp)
{
    unsigned long options;
    spChunkSize nwrite;
    spChunkSize header_size;
    spChunkSize content_size;
    spChunkSize nremain;
    spChunkSize total_nwrite;
    spChunkHeader *parent_header;
    spChunkInfoTable *chunk_info;

    spDebug(80, "spWriteChunk", "in: %c%c%c%c\n",
            chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3]);
    
    if (chunk->parent != NULL) {
        parent_header = &chunk->parent->header;
    } else {
        parent_header = NULL;
    }
    
    chunk_info = spFindChunkInfoTable(spec->info_list, spec->num_list, parent_header, &chunk->header);
    if (chunk_info != NULL) {
        options = chunk_info->options;
    } else {
        options = 0;
    }

    if (rewrite_flag == SP_TRUE && chunk->offset == spTellFile(fp)) {
        spDebug(10, "spWriteChunk", "rewrite_flag == SP_TRUE, offset = %lu\n", (unsigned long)chunk->offset);
        if (options & SP_CHUNK_INFO_OPTION_REWRITE_MASK/* || chunk->child != NULL*/) {
        } else {
            nremain = spec->get_header_size_func(chunk) + spec->get_content_size_func(chunk);
            spSeekFile(fp, nremain, SEEK_CUR);
            spDebug(10, "spWriteChunk", "rewrite not required for %c%c%c%c\n",
                    chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3]);
            return nremain;
        }
    }
    
    chunk->offset = spTellFile(fp);

    if ((nwrite = spec->write_header_func(chunk, &chunk->header, swap, fp)) <= 0) {
        spDebug(10, "spWriteChunk", "write_header_func failed\n");
        return 0;
    }
    spDebug(50, "spWriteChunk", "depth = %ld, nwrite = %lu, chunk->offset = %lu\n", depth, (unsigned long)nwrite, (unsigned long)chunk->offset);

    total_nwrite = nwrite;
    
    if (options && spec->write_header_rest_func != NULL) {
        if ((nwrite = spec->write_header_rest_func(chunk, &chunk->header, options, swap, fp)) <= 0) {
            spDebug(10, "spWriteChunk", "write_header_rest_func failed: nwrite = %lu\n", (unsigned long)nwrite);
        } else {
            total_nwrite += nwrite;
            spDebug(50, "spWriteChunk", "after write_header_rest_func: nwrite = %lu, total_nwrite = %lu\n",
                    (unsigned long)nwrite, (unsigned long)total_nwrite);
        }
    }

    spDebug(50, "spWriteChunk", "write header done: total_nwrite = %lu\n", (unsigned long)total_nwrite);

    if (nwrite >= 0 && chunk_info != NULL) {
        if (options & SP_CHUNK_INFO_OPTION_PAUSE_BEFORE_WRITE_MASK) {
            spDebug(10, "spWriteChunk", "SP_CHUNK_INFO_OPTION_PAUSE_BEFORE_WRITE_MASK found: total_nwrite = %lu\n",
                    (unsigned long)total_nwrite);
            if (paused != NULL) *paused = SP_TRUE;
            return total_nwrite;
        }
        
        if (paused != NULL) *paused = SP_FALSE;
        
        header_size = total_nwrite;
    
        if (options & SP_CHUNK_INFO_OPTION_WRITE_ZEROS_MASK) {
            long len;
            static char buf[8] = {0, 0, 0, 0, 0, 0, 0, 0};

            spDebug(80, "spWriteChunk", "SP_CHUNK_INFO_OPTION_WRITE_ZEROS_MASK found\n");
            
            nremain = spec->get_content_size_func(chunk);
            spDebug(80, "spWriteChunk", "nremain = %lu\n", (unsigned long)nremain);

            while (nremain > 0) {
                len = MIN(8, (long)nremain);
                if ((nwrite = fwrite(buf, 1, len, fp) != len)) {
                    break;
                }

                nremain -= nwrite;
                total_nwrite += nwrite;
            }
        } else {
            if (chunk_info->write_func != NULL) {
                spDebug(80, "spWriteChunk", "call write_func\n");
                if ((nwrite = chunk_info->write_func(chunk, depth, rewrite_flag, swap, fp)) <= 0) {
                    spDebug(10, "spWriteChunk", "calling write_func failed: nwrite = %lu\n", (unsigned long)nwrite);
                } else {
                    total_nwrite += nwrite;
                }
                spDebug(80, "spWriteChunk", "write_func result = %lu\n", (unsigned long)nwrite);
            } 
        }
        
        if (options & SP_CHUNK_INFO_OPTION_PAUSE_AFTER_WRITE_MASK) {
            spDebug(10, "spWriteChunk", "SP_CHUNK_INFO_OPTION_PAUSE_AFTER_WRITE_MASK found: total_nwrite = %lu\n",
                    (unsigned long)total_nwrite);
            if (paused != NULL) *paused = SP_TRUE;
            return total_nwrite;
        }
    }

    header_size = spec->get_header_size_func(chunk);
    content_size = spec->get_content_size_func(chunk);
    spDebug(80, "spWriteChunk", "header_size = %lu, content_size = %lu\n", (unsigned long)header_size, (unsigned long)content_size);
    
    nremain = header_size + content_size - total_nwrite;
    spDebug(80, "spWriteChunk", "total_nwrite = %lu, nremain = %lu, type = %c%c%c%c\n", (unsigned long)total_nwrite, (unsigned long)nremain,
            chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3]);
    
    if (nremain > 0) {
        if (spec->write_content_rest_func != NULL) {
            if ((nwrite = spec->write_content_rest_func(chunk, depth, nremain, rewrite_flag, swap, fp)) <= 0) {
                spDebug(80, "spWriteChunk", "write_content_rest_func failed: nwrite = %lu\n", (unsigned long)nwrite);
            } else {
                total_nwrite += nwrite;
            }
        }
    }
    
    spDebug(80, "spWriteChunk", "done: total_nwrite = %lu, type = %c%c%c%c\n", (unsigned long)total_nwrite,
            chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3]);
    
    return total_nwrite;
}

spChunkSize spWriteChildChunk(spChunkFileSpec *spec, spChunk *chunk, long depth,
                              spBool recursive_flag, spBool rewrite_flag, int swap, spChunk **paused_chunk, FILE *fp)
{
    spChunkSize nwrite;
    spBool paused;
    spChunkSize total_nwrite;
    spChunk *child = NULL;

    total_nwrite = 0;

    if (paused_chunk != NULL && *paused_chunk != NULL) {
        child = (*paused_chunk)->next;
        *paused_chunk = NULL;
    } else {
        child = chunk->child;
    }
        
    while (child != NULL) {
        spDebug(100, "spWriteChildChunk", "total_nwrite = %lu, child type = %c%c%c%c (parent = %c%c%c%c)\n",
                (unsigned long)total_nwrite,
                child->header.type[0], child->header.type[1], child->header.type[2], child->header.type[3],
                chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3]);
        
        paused = SP_FALSE;
        nwrite = spWriteChunk(spec, child, depth + 1, recursive_flag, rewrite_flag, swap, &paused, fp);
        if (nwrite > 0) {
            total_nwrite += nwrite;
        }

        if (paused == SP_TRUE) {
            if (paused_chunk != NULL) *paused_chunk = child;
            break;
        }

        child = child->next;
    }

    spDebug(100, "spWriteChildChunk", "done: total_nwrite = %lu, parent = %c%c%c%c\n",
            (unsigned long)total_nwrite,
            chunk->header.type[0], chunk->header.type[1], chunk->header.type[2], chunk->header.type[3]);
    
    return total_nwrite;
}


spChunkSize spCopyChunk(spChunkFileSpec *spec, spChunk *dest_parent, spChunk *src_chunk, long depth)
{
    spChunkSize ncopy;
    spChunkSize total_ncopy;
    spChunkInfoTable *chunk_info;
    spChunk *dest_chunk;

    chunk_info = spFindChunkInfoTable(spec->info_list, spec->num_list, &dest_parent->header, &src_chunk->header);
    if (chunk_info == NULL) {
        return 0;
    }
    
    dest_chunk = spCreateChunkFromInfoTable(spec, dest_parent, NULL, chunk_info, SP_FALSE, SP_FALSE, NULL, NULL);
    dest_chunk->offset = src_chunk->offset;
    
    if (spec->copy_header_func != NULL) {
        total_ncopy = spec->copy_header_func(&dest_chunk->header, &src_chunk->header);
    } else {
        memcpy(&dest_chunk->header, &src_chunk->header, spec->header_size);
        total_ncopy = spec->header_size;
    }

    if (chunk_info->copy_func != NULL) {
        ncopy = chunk_info->copy_func(dest_chunk, src_chunk, depth);
    } else {
        int offset;

        offset = spOffsetOf(spChunk, header) + spec->header_size;
        ncopy = chunk_info->size - offset;
        if (ncopy > 0) {
            memcpy((char *)dest_chunk + offset, (char *)src_chunk + offset, (long)ncopy);
        }
    }

    if (ncopy > 0) {
        total_ncopy += ncopy;
    }
    
    return total_ncopy;
}

spChunkSize spCopyChildChunk(spChunkFileSpec *spec, spChunk *dest_parent, spChunk *src_parent, long depth)
{
    spChunkSize ncopy;
    spChunkSize total_ncopy;
    spChunk *src_child;

    total_ncopy = 0;
    src_child = src_parent->child;

    while (src_child != NULL) {
        if ((ncopy = spCopyChunk(spec, dest_parent, src_child, depth + 1)) > 0) {
            total_ncopy += ncopy;
        }
        src_child = src_child->next;
    }

    return total_ncopy;
}
